// Copyright (C) 2010-2011 Institut de Robotica i Informatica Industrial, CSIC-UPC.
// Author 
// All rights reserved.
//
// This file is part of iri-ros-pkg
// iri-ros-pkg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 
// IMPORTANT NOTE: This code has been generated through a script from the 
// iri_ros_scripts. Please do NOT delete any comments to guarantee the correctness
// of the scripts. ROS topics can be easly add by using those scripts. Please
// refer to the IRI wiki page for more information:
// http://wikiri.upc.es/index.php/Robotics_Lab

#ifndef _collision_manager_alg_node_h_
#define _collision_manager_alg_node_h_

#include <iri_base_algorithm/iri_base_algorithm.h>
#include "collision_manager_alg.h"
#include <Eigen/Eigen>
#include <tf2/utils.h>

// [publisher subscriber headers]
#include <iri_collision_manager/collision.h>
#include <nav_msgs/Odometry.h>
#include <tf2_ros/transform_broadcaster.h>
#include <geometry_msgs/PoseWithCovarianceStamped.h>
#include <sensor_msgs/LaserScan.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>
#include <tf2_ros/transform_listener.h>
#include <sensor_msgs/Imu.h>

// [service client headers]
#include <iri_laser_scan_icp/icp.h>

// [action server client headers]

/**
 * \struct RangesWithStamp.
 *
 * \brief A struct with the ranges and the time stamp info.
 *
 */
typedef struct {
  std::vector<float> ranges; ///< Laser scan ranges data.
  ros::Time stamp; ///< Laser scan time stamp.
}RangesWithStamp;

/**
 * \struct OdomWithStamp.
 *
 * \brief A struct with the Odometry information and the time stamp.
 *
 */
typedef struct {
  Eigen::Vector3d odom; ///< Odometry data.
  ros::Time stamp; ///< Time stamp.
}OdomWithStamp;

/**
 * \brief IRI ROS Specific Algorithm Class
 *
 */
class CollisionManagerAlgNode : public algorithm_base::IriBaseAlgorithm<CollisionManagerAlgorithm>
{
  private:
    bool new_imu_input_; ///< Flag to know that new input data has been received.
    Eigen::Vector3d imu_local_acc_; ///< Local input linear acceleration.
    Eigen::Vector3d imu_local_ang_vel_; ///< Local input angular speed.
    ros::Time imu_stamp_; ///< IMU stamp.
    ros::Time last_imu_t_; ///< Stamp from the last imu velocity data integrated.
    std::string imu_frame_; ///< IMU frame id.
    std::string front_laser_frame_; ///< Front laser frame id.
    std::string rear_laser_frame_; ///< Rear laser frame id.
    bool tf_base_link_imu_loaded_; ///< Flag to know that tf_base_link_imu has been loaded before.
    Eigen::Transform<double,3,Eigen::Affine> tf_base_link_imu_; ///< Transform from base_link frame to IMU frame.
    bool tf_base_link_front_laser_loaded_; ///< Flag to know that tf_base_link_front_laser has been loaded before.
    tf2::Transform tf_blink_front_laser; ///< Transform from base_link to front laser.
    bool tf_base_link_rear_laser_loaded_; ///< Flag to know that tf_base_link_rear_laser has been loaded before.
    tf2::Transform tf_blink_rear_laser; ///< Transform from base_link to rear laser.
    bool in_collision_; ///< Flag to know that we are receiving collision meassures.
    ros::Time collision_start_stamp_; ///< Time stamp of a collision initial detection.
    // ros::Time collision_end_stamp_; ///< Time stamp at the end of collision.
    Eigen::Vector3d max_acc_; ///< The linear accelerations meassured from base_link frame reference at maximum acceleration module.
    int low_acc_count_; ///< Low acceleration meassures counter to detect the end of a collision.
    int low_ang_vel_count_; ///< Low angular velocities meassures counter to detect the end of a collision.
    std::list<RangesWithStamp> front_ranges_; ///< Front laser ranges buffer.
    std::list<RangesWithStamp> rear_ranges_; ///< Rear laser ranges buffer.
    bool is_front_icp_input_data_ok_; ///<Boolean to know of all the necessary data has been saved on the service request.
    bool is_rear_icp_input_data_ok_; ///<Boolean to know of all the necessary data has been saved on the service request.
    double angular_vel_; ///< Robot's angular velocity from odometry.
    std::string odom_frame_; ///< Odometry Frame id.
    std::list<OdomWithStamp> odom_buffer_; ///< Odometry buffer.
    Eigen::Vector3d odom_before_collision_; ///< Odometry data before collision.
    Eigen::Vector3d odom_after_collision_; ///< Odometry data after collision.
    bool odom_input_is_ok_; ///< Bool to know if the wheel odometry from a collision is ok.

    // [publisher attributes]
    ros::Publisher collisions_publisher_;
    iri_collision_manager::collision collision_msg_;

    ros::Publisher debug_rear_after_icp_scan_publisher_;
    sensor_msgs::LaserScan debug_rear_after_icp_scan_msg_;

    ros::Publisher debug_front_after_icp_scan_publisher_;
    sensor_msgs::LaserScan debug_front_after_icp_scan_msg_;

    tf2_ros::TransformBroadcaster tf2_broadcaster;
    geometry_msgs::TransformStamped transform_msg;

    ros::Publisher debug_rear_icp_pose_publisher_;
    geometry_msgs::PoseWithCovarianceStamped debug_rear_icp_pose_msg_;

    ros::Publisher debug_front_icp_pose_publisher_;
    geometry_msgs::PoseWithCovarianceStamped debug_front_icp_pose_msg_;

    ros::Publisher debug_rear_icp_last_scan_publisher_;
    sensor_msgs::LaserScan debug_rear_icp_last_scan_msg_;

    ros::Publisher debug_rear_icp_ref_scan_publisher_;
    sensor_msgs::LaserScan debug_rear_icp_ref_scan_msg_;

    ros::Publisher debug_front_icp_last_scan_publisher_;
    sensor_msgs::LaserScan debug_front_icp_last_scan_msg_;

    ros::Publisher debug_front_icp_ref_scan_publisher_;
    sensor_msgs::LaserScan debug_front_icp_ref_scan_msg_;


    // [subscriber attributes]
    ros::Subscriber odom_subscriber_;
    void odom_callback(const nav_msgs::Odometry::ConstPtr& msg);
    pthread_mutex_t odom_mutex_;
    void odom_mutex_enter(void);
    void odom_mutex_exit(void);

    ros::Subscriber rear_laser_scan_subscriber_;
    void rear_laser_scan_callback(const sensor_msgs::LaserScan::ConstPtr& msg);
    pthread_mutex_t rear_laser_scan_mutex_;
    void rear_laser_scan_mutex_enter(void);
    void rear_laser_scan_mutex_exit(void);

    ros::Subscriber front_laser_scan_subscriber_;
    void front_laser_scan_callback(const sensor_msgs::LaserScan::ConstPtr& msg);
    pthread_mutex_t front_laser_scan_mutex_;
    void front_laser_scan_mutex_enter(void);
    void front_laser_scan_mutex_exit(void);

    tf2_ros::Buffer tf2_buffer;

    tf2_ros::TransformListener tf2_listener;

    ros::Subscriber imu_subscriber_;
    void imu_callback(const sensor_msgs::Imu::ConstPtr& msg);
    pthread_mutex_t imu_mutex_;
    void imu_mutex_enter(void);
    void imu_mutex_exit(void);


    // [service attributes]

    // [client attributes]
    ros::ServiceClient icp_client_;
    iri_laser_scan_icp::icp front_icp_srv_;
    iri_laser_scan_icp::icp rear_icp_srv_;


    // [action server attributes]

    // [action client attributes]

   /**
    * \brief config variable
    *
    * This variable has all the driver parameters defined in the cfg config file.
    * Is updated everytime function config_update() is called.
    */
    Config config_;
  public:
   /**
    * \brief Constructor
    * 
    * This constructor initializes specific class attributes and all ROS
    * communications variables to enable message exchange.
    */
    CollisionManagerAlgNode(void);

   /**
    * \brief Destructor
    * 
    * This destructor frees all necessary dynamic memory allocated within this
    * this class.
    */
    ~CollisionManagerAlgNode(void);

  protected:
    /**
     * \brief Function to tranform linear local imu data to base_link frame.
     * 
     * It checks if imu frame id is different from base_link frame and if not already 
     * saved, it saves the transfrom.
     * 
     * \param _global_acc The current global linear acceleration.
     * 
     * \param _global_ang_vel The current global angular velocity.
     * 
     * \return True in success.
     */
    bool get_base_link_imu_data(Eigen::Vector3d& _global_acc, Eigen::Vector3d& _global_ang_vel);

    /**
     * \brief Function to check if is an acceleration collision transition.
     * 
     * It checks if the current acceleration meassured is behind or above the threshold  
     * (depending on _start param) updates the counter and checks if the limit has been reached. 
     * 
     * \param _acc_norm_2d The current global linear acceleration norm.
     * 
     * \param _start If it must check the star transition.
     * 
     * \return True if the conditions has been met.
     */
    bool check_acc_collision_transition(double _acc_norm_2d, bool _start);

    /**
     * \brief Function to check if is an acceleration collision transition.
     * 
     * It checks if the current acceleration meassured is behind or above the threshold  
     * (depending on _start param) updates the counter and checks if the limit has been reached. 
     * 
     * \param _ang_vel The current angular velocity.
     * 
     * \param _start If it must check the star transition.
     * 
     * \return True if the conditions has been met.
     */
    bool check_ang_vel_collision_transition(double _ang_vel, bool _start);

    /**
     * \brief Function to check the end of a collision.
     * 
     * It calls check_acc_collision_transition and check_ang_vel_collision_transition
     * with _start to false.
     * 
     * \param _acc_norm_2d The current global linear acceleration norm.
     * 
     * \param _ang_vel The current angular velocity.
     * 
     * \return True if the condition has been met.
     */
    bool check_end_of_collision(double _acc_norm_2d, double _ang_vel);

    /**
     * \brief Function to check the end of a collision.
     * 
     * It calls check_collision_transition with _start to true.
     * 
     * \param _acc_norm_2d The current global linear acceleration norm.
     * 
     * \param _ang_vel The current angular velocity.
     * 
     * \return True if the condition has been met.
     */
    bool check_start_of_collision(double _acc_norm_2d, double _ang_vel);

    /**
     * \brief Function to find the last scan before a collision.
     * 
     * \param _front To select the laser.
     * 
     * \return True if succeeded.
     */
    bool set_ref_ranges(bool _front);

    /**
     * \brief Function to find the last odometry before a collision.
     * 
     * \return True if succeeded.
     */
    bool set_odom_before_collision(void);

    /**
     * \brief Function to find the last scan after a collision.
     * 
     * \param _front To select the laser.
     * 
     * \return True if succeeded.
     */
    bool set_last_ranges(bool _front);

    /**
     * \brief Function to find the last odom after a collision.
     * 
     * \return True if succeeded.
     */
    bool get_last_odom(void);

    /**
     * \brief Function to load tf from base_link to fron/rear laser.
     * 
     * \param _front To select the laser.
     * 
     * \return True if succeeded.
     */
    bool load_tf_blink_laser(bool _front);

    /**
     * \brief Function to get the new robot's base_link pose from ICP result.
     * 
     * It checks that ICP result is good enough and calculates the pose.
     * 
     * \param _new_laser_pose New laser pose from ICP.
     * 
     * \param _covariances Laser pose covariances from ICP.
     * 
     * \param _blink_pose The 2D base link pose.
     * 
     * \param _blink_cov Its covariances.
     * 
     * \param _front To select the laser.
     * 
     * \return True if succeeded.
     */
    bool get_base_link_new_pose_from_icp(const Eigen::Vector3d& _new_laser_pose, const Eigen::Matrix3d& _covariances, Eigen::Vector3d& _blink_pose, Eigen::Matrix3d& _blink_cov, const bool _front);

    /**
     * \brief Function to get the new robot's base_link pose from ICP result.
     * 
     * It checks that ICP result is good enough and calculates the pose.
     * 
     * \param _new_laser_pose New laser pose from ICP.
     * 
     * \param _covariances Laser pose covariances from ICP.
     * 
     * \param _odom The 2D base link pose.
     * 
     * \param _odom_cov Its covariances.
     * 
     * \param _odom_diff The ICP displacment on odom frame.
     * 
     * \param _front To select the laser.
     * 
     * \return True if succeeded.
     */
    bool get_icp_odometry(const Eigen::Vector3d& _new_laser_pose, const Eigen::Matrix3d& _covariances, Eigen::Vector3d& _odom, Eigen::Matrix3d& _odom_cov, Eigen::Vector3d& _odom_diff, const bool _front);

    /**
     * \brief Function to compare the odometry with the ICP to check that is actually a collision.
     * 
     * \param _icp_odometry ICP odometry result.
     * 
     * \param _icp_odometry_diff To return the odometry difference.
     * 
     * \return True if the difference is big enough.
     */
    bool check_odometry_diff(const Eigen::Vector3d& _icp_odometry, Eigen::Vector3d& _icp_odometry_diff);

   /**
    * \brief main node thread
    *
    * This is the main thread node function. Code written here will be executed
    * in every node loop while the algorithm is on running state. Loop frequency 
    * can be tuned by modifying loop_rate attribute.
    *
    * Here data related to the process loop or to ROS topics (mainly data structs
    * related to the MSG and SRV files) must be updated. ROS publisher objects 
    * must publish their data in this process. ROS client servers may also
    * request data to the corresponding server topics.
    */
    void mainNodeThread(void);

   /**
    * \brief dynamic reconfigure server callback
    * 
    * This method is called whenever a new configuration is received through
    * the dynamic reconfigure. The derivated generic algorithm class must 
    * implement it.
    *
    * \param config an object with new configuration from all algorithm 
    *               parameters defined in the config file.
    * \param level  integer referring the level in which the configuration
    *               has been changed.
    */
    void node_config_update(Config &config, uint32_t level);

   /**
    * \brief node add diagnostics
    *
    * In this abstract function additional ROS diagnostics applied to the 
    * specific algorithms may be added.
    */
    void addNodeDiagnostics(void);

    // [diagnostic functions]
    
    // [test functions]
};

#endif
